@@ -18,7 +18,7 @@ from ._version import __version__
 ## Standard library imports
 
 # Generic Operating System Services
-import os, io, time, re
+import os, io, time, re, datetime
 
 # Python Runtime Services
 import traceback
@@ -82,6 +82,7 @@ def update_domain(new_domain):
 domain = get_domain()
 jupyterDrive = ""
 jupyterHome = ""
+userHome = ""
 
 
 jupyter_log = None
@@ -342,11 +343,15 @@ class LaunchNotebookHandler(IPythonHandler):
         try:
             data = json.loads(self.request.body)
             requirementsFile = data.get('requirementsFile')     # we use .get() for params that are not required
-            dataFiles = data.get('dataFiles')                   # we use .get() for params that are not required
+            paths = data.get('paths')
+            hashes = [hash_file(f) for f in paths]
+            stats = [os.stat(f) for f in paths]
+            lastModifieds = [datetime.datetime.fromtimestamp(stat.st_mtime).isoformat() + 'Z' for stat in stats]
+            sizes = [str(stat.st_size) for stat in stats]
             uuid = data['uuid']
             timestamp = data['timestamp']
-            IOLoop.current().run_in_executor(None, launch_notebook, requirementsFile, dataFiles, uuid, timestamp)
-            self.write(json.dumps({'message': 'success'}))
+            IOLoop.current().run_in_executor(None, launch_notebook, requirementsFile, hashes, paths, lastModifieds, sizes, uuid, timestamp)
+            self.write(json.dumps({'message': 'success', 'hashes': hashes, 'files': paths, 'filesmod': lastModifieds, 'filessize': sizes}))
         except Exception as e:
             self.set_status(500)
             self.write(json.dumps({'message': 'Encountered error while launching notebook'}))
@@ -454,8 +459,9 @@ class StopNotebookHandler(IPythonHandler):
     async def post(self):
         try:
             data = json.loads(self.request.body)
-            uuid = data['uuid']
-            response = await IOLoop.current().run_in_executor(None, stop_notebook, uuid)
+            workload = data['workload']
+            module = data['module']
+            response = await IOLoop.current().run_in_executor(None, stop_notebook, workload, module)
             self.set_status(401 if response.geturl().endswith('/login') else response.getcode())
             self.write(response.read())
             if response.getcode() >= 300: jupyter_log.info(optumi_start(self) + str(response))
@@ -476,8 +482,8 @@ class TeardownNotebookHandler(IPythonHandler):
     async def post(self):
         try:
             data = json.loads(self.request.body)
-            uuid = data['uuid']
-            response = await IOLoop.current().run_in_executor(None, teardown_notebook, uuid)
+            workload = data['workload']
+            response = await IOLoop.current().run_in_executor(None, teardown_notebook, workload)
             self.set_status(401 if response.geturl().endswith('/login') else response.getcode())
             self.write(response.read())
             if response.getcode() >= 300: jupyter_log.info(optumi_start(self) + str(response))
@@ -729,8 +735,9 @@ class PullModuleStatusUpdateHandler(IPythonHandler):
             moduleUUIDs = data['moduleUUIDs']
             lastUpdateLines = data['lastUpdateLines']
             lastOutputLines = data['lastOutputLines']
+            lastMonitorings = data.get('lastMonitorings')     # we use .get() for params that are not required
             lastPatches = data['lastPatches']
-            response = await IOLoop.current().run_in_executor(None, pull_module_status_updates, workloadUUIDs, moduleUUIDs, lastUpdateLines, lastOutputLines, lastPatches)
+            response = await IOLoop.current().run_in_executor(None, pull_module_status_updates, workloadUUIDs, moduleUUIDs, lastUpdateLines, lastOutputLines, lastMonitorings, lastPatches)
             self.set_status(401 if response.geturl().endswith('/login') else response.getcode())
             self.write(response.read())
             if response.getcode() >= 300: jupyter_log.info(optumi_start(self) + str(response))
@@ -765,8 +772,9 @@ class DeleteFilesHandler(IPythonHandler):
     async def post(self):
         try:
             data = json.loads(self.request.body)
-            filePaths = data['filePaths']
-            response = await IOLoop.current().run_in_executor(None, delete_files, filePaths)
+            hashes = data['hashes']
+            paths = data['paths']
+            response = await IOLoop.current().run_in_executor(None, delete_files, hashes, paths)
             self.set_status(401 if response.geturl().endswith('/login') else response.getcode())
             self.write(response.read())
             if response.getcode() >= 300: jupyter_log.info(optumi_start(self) + str(response))
@@ -810,9 +818,9 @@ class UploadFilesHandler(IPythonHandler):
         try:
             data = json.loads(self.request.body)
             key = data['key']
-            fileNames = data['fileNames']
+            paths = data['paths']
             compress = data['compress']
-            IOLoop.current().run_in_executor(None, upload_files, key, fileNames, compress)
+            IOLoop.current().run_in_executor(None, upload_files, key, paths, compress)
             self.write(json.dumps({'message': 'success'}))
         except Exception as e:
             self.set_status(500)
@@ -821,6 +829,32 @@ class UploadFilesHandler(IPythonHandler):
             if DEBUG: raise e
 
 class DownloadFilesHandler(IPythonHandler):
+    @authenticated
+    async def post(self):
+        try:
+            data = json.loads(self.request.body)
+            key = data['key']
+            hashes = data['hashes']
+            paths = data['paths']
+            savePaths = data['savePaths']
+            sizes = data['sizes']
+            overwrite = data['overwrite']
+            response = await IOLoop.current().run_in_executor(None, download_files, key, hashes, paths, savePaths, sizes, overwrite)
+            # We only expect a response if something went wrong
+            if response != None: raise response
+        except (ConnectionError, URLError) as e:
+            # If we can't connect to the REST interface, we want the extension to treat it as the user being loggeed out
+            self.set_status(401)
+            self.write(json.dumps({'message': str(e)}))
+            jupyter_log.warning(optumi_start(self) + str(e))
+            if DEBUG: raise e
+        except Exception as e:
+            self.set_status(500)
+            self.write(json.dumps({'message': 'Encountered error while saving notebook output file'}))
+            jupyter_log.error(optumi_start(self) + str(e))
+            if DEBUG: raise e
+
+class GetNotebookOutputFilesHandler(IPythonHandler):
     @authenticated
     async def post(self):
         try:
@@ -828,10 +862,11 @@ class DownloadFilesHandler(IPythonHandler):
             workloadUUID = data['workloadUUID']
             moduleUUID = data['moduleUUID']
             key = data['key']
-            files = data['files']
+            paths = data['paths']
+            savePaths = data['savePaths']
             sizes = data['sizes']
             overwrite = data['overwrite']
-            response = await IOLoop.current().run_in_executor(None, download_files, workloadUUID, moduleUUID, key, files, sizes, overwrite)
+            response = await IOLoop.current().run_in_executor(None, get_notebook_output_files, workloadUUID, moduleUUID, key, paths, savePaths, sizes, overwrite)
             # We only expect a response if something went wrong
             if response != None: raise response
         except (ConnectionError, URLError) as e:
@@ -1070,6 +1105,7 @@ def extract_drive_and_fix_path(path):
 def setup_handlers(server_app):
     global jupyterDrive
     global jupyterHome
+    global userHome
     global jupyter_log
 
     jupyter_log = server_app.log
@@ -1077,6 +1113,7 @@ def setup_handlers(server_app):
     web_app = server_app.web_app
     base_url = web_app.settings['base_url']
     jupyterDrive, jupyterHome = extract_drive_and_fix_path(os.path.expanduser(web_app.settings['server_root_dir']))
+    userHome = extract_drive_and_fix_path(os.path.expanduser('~'))[1]
     host_pattern = '.*$'
     web_app.add_handlers(host_pattern, [(url_path_join(base_url, '/optumi/version'), VersionHandler)])
     web_app.add_handlers(host_pattern, [(url_path_join(base_url, '/optumi/login'), LoginHandler)])
@@ -1107,6 +1144,7 @@ def setup_handlers(server_app):
     web_app.add_handlers(host_pattern, [(url_path_join(base_url, '/optumi/delete-files'), DeleteFilesHandler)])
     web_app.add_handlers(host_pattern, [(url_path_join(base_url, '/optumi/list-files'), ListFilesHandler)])
     web_app.add_handlers(host_pattern, [(url_path_join(base_url, '/optumi/download-files'), DownloadFilesHandler)])
+    web_app.add_handlers(host_pattern, [(url_path_join(base_url, '/optumi/get-notebook-output-files'), GetNotebookOutputFilesHandler)])
     web_app.add_handlers(host_pattern, [(url_path_join(base_url, '/optumi/get-download-progress'), GetDownloadProgressHandler)])
     web_app.add_handlers(host_pattern, [(url_path_join(base_url, '/optumi/get-total-billing'), GetTotalBillingHandler)])
     web_app.add_handlers(host_pattern, [(url_path_join(base_url, '/optumi/get-detailed-billing'), GetDetailedBillingHandler)])
@@ -1173,7 +1211,7 @@ def logout():
         finally:
             lock.release()
         req = request.Request(URL)
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1191,7 +1229,7 @@ def sign_agreement(timeOfSigning, hashOfSignedAgreement):
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1199,7 +1237,7 @@ def get_new_agreement():
     URL = get_path() + '/exp/jupyterlab/get-new-agreement'
     try:
         req = request.Request(URL)
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1212,7 +1250,7 @@ def exchange_versions(version):
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1220,7 +1258,7 @@ def get_user_information():
     URL = get_path() + '/exp/jupyterlab/get-user-information'
     try:
         req = request.Request(URL)
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1235,7 +1273,7 @@ def set_user_information(param, value):
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1262,9 +1300,9 @@ def preview_notebook(nbConfig):
 
         # jupyter_log.info(optumi_start() + )
         # jupyter_log.info(optumi_start() + 'SERVER RESPONSE:')
-        # jupyter_log.info(optumi_start() + request.urlopen(req).read().decode('utf-8'))
+        # jupyter_log.info(optumi_start() + request.urlopen(req, timeout=120).read().decode('utf-8'))
 
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1282,21 +1320,24 @@ def setup_notebook(name, timestamp, notebook, nbConfig):
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
-def launch_notebook(requirementsFile, dataFiles, uuid, timestamp):
+def launch_notebook(requirementsFile, hashes, paths, lastModifieds, sizes, uuid, timestamp):
     URL = get_path() + '/exp/jupyterlab/launch-notebook'
     try:
         form = MultiPartForm()
         form.add_field('uuid', uuid)
         ## Send this list as a file since it can have arbitrary length and cause issues for the REST Interface if sent as an array
-        form.add_file('dataFileNames', 'dataFileNames', fileHandle=io.BytesIO(','.join(dataFiles).encode('utf-8')))
+        form.add_file('hashes', 'hashes', fileHandle=io.BytesIO(','.join(hashes).encode('utf-8')))
+        form.add_file('paths', 'paths', fileHandle=io.BytesIO(','.join(paths).encode('utf-8')))
+        form.add_file('lastModifieds', 'lastModifieds', fileHandle=io.BytesIO(','.join(lastModifieds).encode('utf-8')))
+        form.add_file('sizes', 'sizes', fileHandle=io.BytesIO(','.join(sizes).encode('utf-8')))
         form.add_field('timestamp', timestamp)
         form.add_field('jupyterDrive', jupyterDrive)
         form.add_field('jupyterHome', jupyterHome)
-        form.add_field('userHome', extract_drive_and_fix_path(os.path.expanduser('~'))[1])
+        form.add_field('userHome', userHome)
 
         try:
             lock.acquire()
@@ -1317,7 +1358,7 @@ def launch_notebook(requirementsFile, dataFiles, uuid, timestamp):
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        response = request.urlopen(req)
+        response = request.urlopen(req, timeout=120)
         try:
             lock.acquire()
             if response.getcode() == 200:
@@ -1337,11 +1378,12 @@ def launch_notebook(requirementsFile, dataFiles, uuid, timestamp):
             lock.release()
         raise
 
-def stop_notebook(uuid):
+def stop_notebook(workload, module):
     URL = get_path() + '/exp/jupyterlab/stop-notebook'
     try:
         form = MultiPartForm()
-        form.add_field('uuid', uuid)
+        form.add_field('workload', workload)
+        form.add_field('module', module)
         data = bytes(form)
 
         try:
@@ -1355,15 +1397,15 @@ def stop_notebook(uuid):
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
-def teardown_notebook(uuid):
+def teardown_notebook(workload):
     URL = get_path() + '/exp/jupyterlab/teardown-notebook'
     try:
         form = MultiPartForm()
-        form.add_field('uuid', uuid)
+        form.add_field('workload', workload)
         data = bytes(form)
 
         try:
@@ -1377,12 +1419,12 @@ def teardown_notebook(uuid):
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
 def upload_cleanup(key, pairs):
-    for response, fileName in pairs:
+    for response, path in pairs:
         response.read()
         try:
             lock.acquire()
@@ -1391,25 +1433,25 @@ def upload_cleanup(key, pairs):
             pass
         finally:
             lock.release()
-        if fileName != None:
+        if path != None:
             try:
-                os.remove(fileName)
+                os.remove(path)
             except Exception as e:
-                print(e)
+                pass
 
-def upload_files(key, fileNames, compress):
+def upload_files(key, paths, compress):
     URL = get_path() + '/exp/jupyterlab/upload-files'
     try:
         filesToUpload = []
-        ## Check for files in chunks of 100
-        while len(fileNames) > 100:
-            chunk = fileNames[:100]
+        ## Check for files in chunks of 1000
+        while len(paths) > 1000:
+            chunk = paths[:1000]
             for exists, file in zip(json.load(check_if_files_exist(chunk))['exists'], chunk):
                 if not exists:
                     filesToUpload.append((file, os.path.getsize(file)))
-            fileNames = fileNames[100:]
+            paths = paths[1000:]
         ## Check last chunk
-        for exists, file in zip(json.load(check_if_files_exist(fileNames))['exists'], fileNames):
+        for exists, file in zip(json.load(check_if_files_exist(paths))['exists'], paths):
             if not exists:
                 filesToUpload.append((file, os.path.getsize(file)))
         
@@ -1422,12 +1464,11 @@ def upload_files(key, fileNames, compress):
                     lock.acquire()
                     compressionProgress[key] = {'progress': 0, 'total': len(filesToUpload)}
                 except Exception as e:
-                    print(e)
                     pass
                 finally:
                     lock.release()
 
-            MAX_SIZE = 5 * 1024 * 1024 * 1024
+            MAX_SIZE = 2 * 1024 * 1024 * 1024
 
             fileChunks = [[]]
             totalSize = 0
@@ -1462,7 +1503,6 @@ def upload_files(key, fileNames, compress):
                     compressionProgress[key]['progress'] = -1
                 uploadProgress[key] = { 'progress': 0, 'total': totalSize }
             except Exception as e:
-                print(e)
                 pass
             finally:
                 lock.release()
@@ -1474,6 +1514,7 @@ def upload_files(key, fileNames, compress):
                     form.add_file('files', file, fileHandle=open(file, 'rb'))
 
                 form.add_field('jupyterHome', jupyterHome)
+                form.add_field('userHome', userHome)
                 form.add_field('compressed', str(compress))
 
                 b = bytes(form)
@@ -1482,7 +1523,7 @@ def upload_files(key, fileNames, compress):
                 req = request.Request(URL, data=data)
                 req.add_header('Content-type', form.get_content_type())
                 req.add_header('Content-length', len(data))
-                res = request.urlopen(req)
+                res = request.urlopen(req, timeout=120)
                 if compress:
                     responses.append((res, chunk[0]))
                 else:
@@ -1498,11 +1539,10 @@ def upload_files(key, fileNames, compress):
                     compressionProgress[key] = {'progress': -1, 'total': 0 }
                 uploadProgress[key] = {'progress': -1, 'total': 0 }
             except Exception as e:
-                print(e)
                 pass
             finally:
                 lock.release()
-    except HTTPError as e:
+    except (CancelledError, HTTPError) as e:
         # Clean up ongoing progress in the case of a failure
         try:
             lock.acquire()
@@ -1514,27 +1554,28 @@ def upload_files(key, fileNames, compress):
             if key in uploadProgress:
                 uploadProgress[key] = {'progress': -1, 'total': 0 }
         except Exception as e:
-            print(e)
             pass
         finally:
             lock.release()
 
         return e
 
-def check_if_files_exist(fileNames):
+def check_if_files_exist(paths):
     URL = get_path() + '/exp/jupyterlab/check-if-files-exist'
     try:
         form = MultiPartForm()
-        for file in fileNames:
-            form.add_field('fileNames', file)
-            form.add_field('hashes', hash_file(file))
+        ## Send this list as a file since it can have arbitrary length and cause issues for the REST Interface if sent as an array
+        form.add_file('paths', 'paths', fileHandle=io.BytesIO(','.join(paths).encode('utf-8')))
+        form.add_file('hashes', 'hashes', fileHandle=io.BytesIO(','.join([hash_file(f) for f in paths]).encode('utf-8')))
+        form.add_field('userHome', jupyterHome)
         form.add_field('jupyterHome', jupyterHome)
+        form.add_field('autoAdd', str(True))
         data = bytes(form)
 
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1542,7 +1583,7 @@ def get_machines():
     URL = get_path() + '/exp/jupyterlab/get-machines'
     try:
         req = request.Request(URL)
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1556,7 +1597,7 @@ def pull_workload_config(workload):
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1571,7 +1612,7 @@ def push_workload_config(workload, nbConfig):
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1585,7 +1626,7 @@ def get_notebook_config(nbKey):
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1600,7 +1641,7 @@ def set_notebook_config(nbKey, nbConfig):
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1608,7 +1649,7 @@ def get_data_connectors():
     URL = get_path() + '/exp/jupyterlab/get-data-connectors'
     try:
         req = request.Request(URL)
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1624,7 +1665,7 @@ def add_data_connector(dataService, name, info):
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1638,7 +1679,7 @@ def remove_data_connector(name):
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1654,7 +1695,7 @@ def push_workload_status_update(uuid, phase, update):
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1675,11 +1716,11 @@ def pull_workload_status_updates(uuids, lastInitializingLines, lastPreparingLine
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
-def pull_module_status_updates(workloadUUIDs, moduleUUIDs, lastUpdateLines, lastOutputLines, lastPatches):
+def pull_module_status_updates(workloadUUIDs, moduleUUIDs, lastUpdateLines, lastOutputLines, lastMonitorings, lastPatches):
     URL = get_path() + '/exp/jupyterlab/pull-module-status-updates'
     try:
         form = MultiPartForm()
@@ -1691,6 +1732,9 @@ def pull_module_status_updates(workloadUUIDs, moduleUUIDs, lastUpdateLines, last
             form.add_field('lastUpdateLines', str(lastUpdateLine))
         for lastOutputLine in lastOutputLines:
             form.add_field('lastOutputLines', str(lastOutputLine))
+        if lastMonitorings != None:
+            for lastMonitoring in lastMonitorings:
+                form.add_field('lastMonitorings', str(lastMonitoring))
         for lastPatch in lastPatches:
             form.add_field('lastPatches', str(lastPatch))
         data = bytes(form)
@@ -1698,7 +1742,7 @@ def pull_module_status_updates(workloadUUIDs, moduleUUIDs, lastUpdateLines, last
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1706,27 +1750,27 @@ def list_files():
     URL = get_path() + '/exp/jupyterlab/list-files'
     try:
         req = request.Request(URL)
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
-def delete_files(filePaths):
+def delete_files(hashes, paths):
     URL = get_path() + '/exp/jupyterlab/delete-files'
     try:
         form = MultiPartForm()
-        for filePath in filePaths:
-            form.add_field('filePaths', str(filePath))
+        form.add_file('hashes', 'hashes', fileHandle=io.BytesIO(','.join(hashes).encode('utf-8')))
+        form.add_file('paths', 'paths', fileHandle=io.BytesIO(','.join(paths).encode('utf-8')))
         data = bytes(form)
 
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
 def save_file(fileName, content, overwrite):
-    file = fileName
+    file = os.path.expanduser(fileName)
     if not overwrite:
         newName = file
         num = 1
@@ -1741,7 +1785,7 @@ def save_file(fileName, content, overwrite):
     with open(file, "wb") as f:
         f.write(base64.decodebytes(content.getvalue()))
 
-def add_content(content, decoded):
+def add_content(key, content, decoded):
     # Remove starting or ending " so we can look at only , when splitting the string
     if decoded != '' and decoded[0] == '"':
         decoded = decoded[1:]
@@ -1751,22 +1795,20 @@ def add_content(content, decoded):
     content.write(encoded)
     try:
         lock.acquire()
-        downloadProgress[key]['progress'] = downloadProgress[key]['progress'] + len(encoded)
+        downloadProgress[key]['progress'] = downloadProgress[key]['progress'] + len(base64.decodebytes(encoded))
     except:
         pass
     finally:
         lock.release()
 
-def download_files_helper(key, form, fileNames, overwrite):
-    URL = get_path() + '/exp/jupyterlab/get-notebook-output-file'
-
+def download_files_helper(URL, key, form, savePaths, overwrite):
     data = bytes(form)
 
     req = request.Request(URL, data=data)
     req.add_header('Content-type', form.get_content_type())
     req.add_header('Content-length', len(data))
 
-    response = request.urlopen(req)
+    response = request.urlopen(req, timeout=120)
     total = response.getheader('content-length')
 
     blocksize = 4096
@@ -1786,18 +1828,20 @@ def download_files_helper(key, form, fileNames, overwrite):
         if ',' in decoded:
             split = decoded.split(',')
             for chunk in split[:-1]:
-                add_content(content, chunk)
-                save_file(fileNames.pop(0), content, overwrite)
+                add_content(key, content, chunk)
+                save_file(savePaths.pop(0), content, overwrite)
                 content = io.BytesIO()
             # Handle the last chunk like the else case below
-            add_content(content, split[-1])
+            add_content(key, content, split[-1])
         else:
-            add_content(content, decoded)
+            add_content(key, content, decoded)
 
     # Save the last file
-    save_file(fileNames.pop(0), content, overwrite)
+    save_file(savePaths.pop(0), content, overwrite)
+
+def download_files(key, hashes, paths, savePaths, sizes, overwrite):
+    URL = get_path() + '/exp/jupyterlab/download-files'
 
-def download_files(workloadUUID, moduleUUID, key, files, sizes, overwrite):
     try:
         total = sum(sizes)
         try:
@@ -1808,37 +1852,100 @@ def download_files(workloadUUID, moduleUUID, key, files, sizes, overwrite):
         finally:
             lock.release()
 
-        filesToDownload = zip(files, sizes)
+        filesToDownload = zip(hashes, paths, savePaths, sizes)
         # Sort by file size
-        filesToDownload = sorted(filesToDownload, key=lambda x:x[1], reverse=True)
+        filesToDownload = sorted(filesToDownload, key=lambda x:x[3], reverse=True)
 
-        MAX_SIZE = 5 * 1024 * 1024 * 1024
+        MAX_SIZE = 2 * 1024 * 1024 * 1024
+
+        totalSize = 0
+        form = MultiPartForm()
+        formHashes = []
+        fileNames = []
+        saveNames = []
+        for hash, file, savePath, size in filesToDownload:
+            if size > MAX_SIZE:
+                # TODO:JJ We should let the user know we can;t upload a file because it is too large
+                continue
+            elif totalSize + size > MAX_SIZE:
+                form.add_file('hashes', 'hashes', fileHandle=io.BytesIO(','.join(formHashes).encode('utf-8')))
+                form.add_file('paths', 'paths', fileHandle=io.BytesIO(','.join(fileNames).encode('utf-8')))
+                download_files_helper(URL, key, form, saveNames, overwrite)
+
+                # Reset variables for the next chunk of files
+                totalSize = 0
+                form = MultiPartForm()
+                formHashes = []
+                fileNames = []
+                saveNames = []
+            formHashes.append(hash)
+            fileNames.append(file)
+            saveNames.append(savePath)
+        # Download the remaining files
+        form.add_file('hashes', 'hashes', fileHandle=io.BytesIO(','.join(formHashes).encode('utf-8')))
+        form.add_file('paths', 'paths', fileHandle=io.BytesIO(','.join(fileNames).encode('utf-8')))
+        download_files_helper(URL, key, form, saveNames, overwrite)
+
+        # Make sure we mark the download as completed
+        try:
+            lock.acquire()
+            downloadProgress[key] = { 'progress': -1, 'total': total }
+        except:
+            pass
+        finally:
+            lock.release()
+
+        return
+    except HTTPError as e:
+        return e
+
+def get_notebook_output_files(workloadUUID, moduleUUID, key, paths, savePaths, sizes, overwrite):
+    URL = get_path() + '/exp/jupyterlab/get-notebook-output-file'
+
+    try:
+        total = sum(sizes)
+        try:
+            lock.acquire()
+            downloadProgress[key] = { 'progress': 0, 'total': total }
+        except:
+            pass
+        finally:
+            lock.release()
+
+        filesToDownload = zip(paths, savePaths, sizes)
+        # Sort by file size
+        filesToDownload = sorted(filesToDownload, key=lambda x:x[2], reverse=True)
+
+        MAX_SIZE = 2 * 1024 * 1024 * 1024
 
         totalSize = 0
         form = MultiPartForm()
         fileNames = []
-        for file, size in filesToDownload:
+        saveNames = []
+        for file, savePath, size in filesToDownload:
             if size > MAX_SIZE:
                 # TODO:JJ We should let the user know we can;t upload a file because it is too large
                 continue
             elif totalSize + size > MAX_SIZE:
                 form.add_field('workloadUUID', workloadUUID)
                 form.add_field('moduleUUID', moduleUUID)
+                form.add_file('paths', 'paths', fileHandle=io.BytesIO(','.join(fileNames).encode('utf-8')))
 
-                download_files_helper(key, form, fileNames, overwrite)
+                download_files_helper(URL, key, form, saveNames, overwrite)
 
                 # Reset variables for the next chunk of files
                 totalSize = 0
                 form = MultiPartForm()
                 fileNames = []
-            else: 
-                form.add_field('fileNames', file)
-                fileNames.append(file)
+                saveNames = []
+            fileNames.append(file)
+            saveNames.append(savePath)
         # Download the remaining files
         form.add_field('workloadUUID', workloadUUID)
         form.add_field('moduleUUID', moduleUUID)
+        form.add_file('paths', 'paths', fileHandle=io.BytesIO(','.join(fileNames).encode('utf-8')))
 
-        download_files_helper(key, form, fileNames, overwrite)
+        download_files_helper(URL, key, form, saveNames, overwrite)
 
         # Make sure we mark the download as completed
         try:
@@ -1864,7 +1971,7 @@ def get_total_billing(startTime, endTime):
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1879,7 +1986,7 @@ def get_detailed_billing(startTime, endTime):
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1893,7 +2000,7 @@ def delete_machine(uuid):
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1909,7 +2016,7 @@ def change_password(loginName, oldPassword, newPassword):
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1925,7 +2032,7 @@ def create_checkout(items, redirect):
         req = request.Request(URL, data=data)
         req.add_header('Content-type', form.get_content_type())
         req.add_header('Content-length', len(data))
-        return request.urlopen(req)
+        return request.urlopen(req, timeout=120)
     except HTTPError as e:
         return e
 
@@ -1941,7 +2048,7 @@ def hash_file(fileName):
         while len(buf) > 0:
             hasher.update(buf)
             buf = afile.read(BLOCKSIZE)
-    return hasher.hexdigest()
+    return hasher.hexdigest().upper()
 
 ###############################################################################################
 ####    Zip files
