@@ -22,7 +22,7 @@ export enum StatusColor {
     RED = '#f48f8d',
     ORANGE = '#ffab61',
     YELLOW = '#fff21c',
-    GREEN = '#68da7c',
+    GREEN = '#7cdf8e',
     BLUE = '#10A0F9',
     LIGHT_BLUE = '#c3d7ff',
     PURPLE = '#934692',
