@@ -480,7 +480,6 @@ class FilesPanel extends React.Component<IProps, IState> {
     private refreshFiles = async () => {
         if (this.refreshingFiles) {
             if (Global.shouldLogOnPoll) console.log('FunctionPoll (' + new Date().getSeconds() + ')');
-            setTimeout(this.refreshFiles, 60000);
             const optumi = Global.metadata.getMetadata().config;
             const files = optumi.upload.files;
             for (var file of files) {
@@ -493,13 +492,13 @@ class FilesPanel extends React.Component<IProps, IState> {
                     }
                 }
             }
+            setTimeout(this.refreshFiles, 60000);
         }
     }
 
     private refreshDataConnectors = async () => {
         if (this.refreshingDataConnectors) {
             if (Global.shouldLogOnPoll) console.log('FunctionPoll (' + new Date().getSeconds() + ')');
-            setTimeout(this.refreshDataConnectors, 60000);
             const optumi = Global.metadata.getMetadata().config;
             const dataConnectors = optumi.upload.dataConnectors;
             
@@ -519,9 +518,12 @@ class FilesPanel extends React.Component<IProps, IState> {
                     }
                 }
             }
+            setTimeout(this.refreshDataConnectors, 60000);
         }
     }
 
+    private handleDataConnectorChange = () => this.refreshDataConnectors()
+
     private handleMetadataChange = () => {this.forceUpdate()}
 	private handleLabShellChange = () => {this.forceUpdate()}
 
@@ -534,12 +536,14 @@ class FilesPanel extends React.Component<IProps, IState> {
         this.refreshDataConnectors();
 		Global.metadata.getMetadataChanged().connect(this.handleMetadataChange);
 		Global.labShell.currentChanged.connect(this.handleLabShellChange);
+        Global.dataConnectorChange.connect(this.handleDataConnectorChange);
 	}
 
 	// Will be called automatically when the component is unmounted
 	public componentWillUnmount = () => {
         Global.metadata.getMetadataChanged().disconnect(this.handleMetadataChange);
         Global.labShell.currentChanged.disconnect(this.handleLabShellChange);
+        Global.dataConnectorChange.disconnect(this.handleDataConnectorChange);
         this.refreshingFiles = false;
         this.refreshingDataConnectors = false;
         this._isMounted = false
@@ -643,13 +647,13 @@ class ResourceFile extends React.Component<RFProps, RFState> {
                         display: 'inline-flex'
                     }}
                 >
-                    {(compression.length > 0 || (upload.length > 0 && upload[0].total < 0)) ? (
+                    {(!this.props.noLongerExists && (compression.length > 0 || (upload.length > 0 && upload[0].total < 0))) ? (
                         <ExtraInfo reminder={compression.length > 0 ? compression[0].total == -1 ? '' : 'Compressed ' + compression[0].progress + '/' + compression[0].total + ' files' : ''}>
                             <div style={{height: '16px', width: '16px', marginRight: '4px'}}>
                                 <CircularProgress color="primary" size='14px' thickness={8} style={{margin: 'auto'}}/>
                             </div>
                         </ExtraInfo>
-                    ) : (upload.length > 0 ? (
+                    ) : (!this.props.noLongerExists && (upload.length > 0) ? (
                         <ExtraInfo reminder={FormatUtils.styleCapacityUnitValue()(upload[0].progress) + '/' + FormatUtils.styleCapacityUnitValue()(upload[0].total)}>
                             <div style={{height: '16px', width: '16px', marginRight: '4px'}}>
                                 <CircularProgress variant="determinate" size='14px' thickness={8} style={{margin: 'auto'}} value={(upload[0].progress / upload[0].total) * 100 } />
