@@ -26,7 +26,9 @@ import { WasabiConnectorPopup } from '../deploy/WasabiConnectorPopup';
 import { PhoneTextBox } from '../../core/PhoneTextBox';
 import { AzureBlobStorageConnectorPopup } from '../deploy/AzureBlobStorageConnector';
 import DeleteIcon from '@material-ui/icons/Delete';
-import ExtraInfo from '../../utils/ExtraInfo';
+import { FileMetadata } from '../deploy/fileBrowser/FileBrowser';
+import { FileTree } from '../FileTree';
+import moment from 'moment';
 
 // Properties from parent
 interface IProps {
@@ -115,11 +117,91 @@ export class AccountGeneralSubMenu extends React.Component<IProps, IAccountGener
                         labelBefore
                         flip
                     />
+                    {Global.user.showMonitoringEnabled && (
+                        <Button 
+                            variant="outlined"
+                            color="primary"
+                            startIcon={<GetAppIcon />}
+                            style={{width: '100%'}}
+                            onClick={async () => {
+                                const data = [];
+                                for (let app of Global.user.appTracker.finishedJobsOrSessions) {
+                                    const machine = app.machine;
+                                    data.push(
+                                        [
+                                            this.stringToCSVCell(app.name),
+                                            this.stringToCSVCell(app.annotationOrRunNum),
+                                            app.timestamp,
+                                            app.getTimeElapsed(),
+                                            app.getCost(),
+                                            app.getAppMessage(),
+                                            machine.name,
+                                            machine.computeCores,
+                                            FormatUtils.styleCapacityUnitValue()(machine.memorySize),
+                                            FormatUtils.styleCapacityUnitValue()(machine.storageSize),
+                                            machine.graphicsNumCards > 0 ? (machine.graphicsNumCards + ' ' + machine.graphicsCardType) : 'None',
+                                            app.uuid,
+                                        ]
+                                    );
+                                    
+                                    var link = document.createElement("a");
+                                    var blob = new Blob([JSON.stringify(app.machine)], {
+                                        type: "text/plain;charset=utf-8"
+                                    });
+                                    link.setAttribute("href", window.URL.createObjectURL(blob));
+                                    link.setAttribute("download", app.uuid + ".txt");
+                                    document.body.appendChild(link); // Required for FF
+                                    link.click();
+
+                                    await new Promise(resolve => setTimeout(resolve, 100));
+
+                                    // const ipynbContent = 'data:text/plain;charset=utf-8,'
+                                    //     + JSON.stringify(app.notebook);
+                                    var link = document.createElement("a");
+                                    var blob = new Blob([JSON.stringify(app.notebook)], {
+                                        type: "text/plain;charset=utf-8"
+                                    });
+                                    link.setAttribute("href", window.URL.createObjectURL(blob));
+                                    link.setAttribute("download", app.uuid + ".ipynb");
+                                    document.body.appendChild(link); // Required for FF
+                                    link.click();
+
+                                    await new Promise(resolve => setTimeout(resolve, 100));
+                                }
+            
+                                const headers = [
+                                    ["Name", "Annotation", "Start Time", "Duration", "Cost", "Status", "Machine", "Cores", "RAM", "Disk", "GPUs", "UUID"]
+                                ];
+            
+                                var link = document.createElement("a");
+                                var blob = new Blob([headers.map(e => e.join(",")).join("\n") + '\n' + data.map(e => e.join(",")).join("\n") + '\n'], {
+                                    type: "data:text/csv;charset=utf-8,"
+                                });
+                                link.setAttribute("href", window.URL.createObjectURL(blob));
+                                link.setAttribute("download", "run_history.csv");
+                                document.body.appendChild(link); // Required for FF
+                                link.click();
+                            }
+                        }>
+                            Download all runs
+                        </Button>
+                    )}
                 </div>
             </>
         )
     }
 
+    public stringToCSVCell(str: string): string {
+        var s = "\"";
+        for(let nextChar of str) {
+            s += nextChar;
+            if (nextChar == '"')
+                s += "\"";
+        }
+        s += "\"";
+        return s;
+    }
+
     public shouldComponentUpdate = (nextProps: IProps, nextState: IState): boolean => {
         try {
             if (JSON.stringify(this.props) != JSON.stringify(nextProps)) return true;
@@ -428,14 +510,12 @@ export class AccountPaymentSubMenu extends React.Component<IProps, IAccountPayme
                     const headers = [
                         ["Start Time", "End Time", "Duration", "Rate ($/hr)", "Cost", "GPU", "CPU", "RAM", "Disk"]
                     ];
-                                        
-                    const csvContent = "data:text/csv;charset=utf-8," 
-                        + headers.map(e => e.join(",")).join("\n") + '\n'
-                        + data.map(e => e.join(",")).join("\n") + '\n';
-
-                    var encodedUri = encodeURI(csvContent);
+                    
                     var link = document.createElement("a");
-                    link.setAttribute("href", encodedUri);
+                    var blob = new Blob([headers.map(e => e.join(",")).join("\n") + '\n' + data.map(e => e.join(",")).join("\n") + '\n'], {
+                        type: "data:text/csv;charset=utf-8,"
+                    });
+                    link.setAttribute("href", window.URL.createObjectURL(blob));
                     link.setAttribute("download", "billing_records.csv");
                     document.body.appendChild(link); // Required for FF
                     link.click();
@@ -556,7 +636,7 @@ export class AccountConnectorsSubMenu extends React.Component<IProps, IAccountCo
                                 newDataConnectors = newDataConnectors.filter(dataConnector => dataConnector.name !== dataConnectorMetadata.name)
                                 this.safeSetState({dataConnectors: newDataConnectors})
                                 this.forceNewBrowser()
-                            });                            
+                            }).then(() => Global.dataConnectorChange.emit(void 0));                   
                         }}
                     />
                     <Divider style={{marginTop: '33px'}}/>
@@ -619,131 +699,150 @@ export class AccountConnectorsSubMenu extends React.Component<IProps, IAccountCo
     }
 }
 
-
-interface IAccountFilesSubMenuState {
-    files: string[];
-}
+interface IAccountFilesSubMenuState {}
 
 export class AccountFilesSubMenu extends React.Component<IProps, IAccountFilesSubMenuState> {
-    private _isMounted = false;
-    private polling = false;
-
     constructor(props: IProps) {
         super(props);
         this.state = {
             files: [],
+            appsToFiles: new Map(),
+            filesToApps: new Map(),
         }
     }
 
     public render = (): JSX.Element => {
 		if (Global.shouldLogOnRender) console.log('ComponentRender (' + new Date().getSeconds() + ')');
+        const files = Global.user.fileTracker.files;
         return (
             <>
                 <div style={Object.assign({padding: '20px 6px 6px 20px'}, this.props.style)}>
                     <div>
                         Listed files are securely stored in the Optumi platform. If you want to delete a file hit its associated trash can icon.
+                        {this.getTotalUsedStorage() > 0 && <>
+                            <br />
+                            <br />
+                            Total used storage: {this.formatSize(this.getTotalUsedStorage())}
+                        </>}
                     </div>
-                    {this.state.files.map(x => (
-
-                        <div style={{display: 'inline-flex', width: '100%'}}>
-                            <ExtraInfo reminder={x}>
-                                <div style={{margin: 'auto'}}>
-                                    {x.split('/').pop()}
-                                </div>
-                            </ExtraInfo>
-                            <div style={{flexGrow: 1}}/>
-                            <IconButton
-                                onClick={() => this.deleteFile(x)}
-                                style={{position: 'relative', display: 'inline-block', width: '36px', height: '36px', padding: '3px'}}
-                            >
-                                <DeleteIcon style={{position: 'relative', width: '30px', height: '30px', padding: '3px'}} />
-                            </IconButton>
-                        </div>
-                    ))}
+                    <FileTree<FileMetadata>
+                        files={files}
+                        fileTitle={file => (
+`Name: ${file.name}
+${file.size === null ? '' : `Size: ${this.formatSize(file.size)}
+`}${file.path === '' ? '' : `Path: ${file.path.replace(file.name, '').replace(/\/$/, '')}
+`}Modified: ${moment(file.last_modified).format('YYYY-MM-DD hh:mm:ss')}`
+                        )}
+                        fileHidableIcon={file => ({
+                            width: 72,
+                            height: 36,
+                            icon: (
+                                <>
+                                    <IconButton
+                                        onClick={() => this.downloadFile(file)}
+                                        style={{ width: '36px', height: '36px', padding: '3px' }}
+                                    >
+                                        <GetAppIcon style={{ width: '30px', height: '30px', padding: '3px' }} />
+                                    </IconButton>
+                                    <IconButton
+                                        onClick={() => this.deleteFile(file)}
+                                        style={{ width: '36px', height: '36px', padding: '3px' }}
+                                    >
+                                        <DeleteIcon style={{ width: '30px', height: '30px', padding: '3px' }} />
+                                    </IconButton>
+                                </>
+                            ),
+                        })}
+                        directoryHidableIcon={path => ({
+                            width: 72,
+                            height: 36,
+                            icon: (
+                                <>
+                                    <IconButton
+                                        onClick={() => this.downloadDirectory(path)}
+                                        style={{ width: '36px', height: '36px', padding: '3px' }}
+                                    >
+                                        <GetAppIcon style={{ width: '30px', height: '30px', padding: '3px' }} />
+                                    </IconButton>
+                                    <IconButton
+                                        onClick={() => this.deleteDirectory(path)}
+                                        style={{ width: '36px', height: '36px', padding: '3px' }}
+                                    >
+                                        <DeleteIcon style={{ width: '30px', height: '30px', padding: '3px' }} />
+                                    </IconButton>
+                                </>
+                            ),
+                        })}
+                    />
                 </div>
             </>
         )
     }
 
-    private async deleteFile(filePath: string) {
-		const settings = ServerConnection.makeSettings();
-        const url = settings.baseUrl + "optumi/delete-files";
-		const init: RequestInit = {
-			method: 'POST',
-			body: JSON.stringify({
-                filePaths: [filePath],
-            }),
-		};
-		ServerConnection.makeRequest(
-			url,
-			init, 
-			settings
-		).then((response: Response) => {
-			Global.handleResponse(response);
-			return response.json();
-		});
-        this.safeSetState({ files: this.state.files.filter(x => x != filePath) });
+    private formatSize = (value: number) => {
+		if (value == 0) return "";
+		if (value < Math.pow(1024, 1)) {
+            return value.toFixed() + ' B';
+        } else if (value < Math.pow(1024, 2)) {
+            return (value / Math.pow(1024, 1)).toFixed(1) + ' KiB';
+        } else if (value < Math.pow(1024, 3)) {
+            return (value / Math.pow(1024, 2)).toFixed(1) + ' MiB';
+        } else if (value < Math.pow(1024, 4)) {
+            return (value / Math.pow(1024, 3)).toFixed(1) + ' GiB';
+        } else if (value < Math.pow(1024, 5)) {
+            return (value / Math.pow(1024, 4)).toFixed(1) + ' TiB';
+        } else {
+            return (value / Math.pow(1024, 5)).toFixed(1) + ' PiB';
+        }
+	}
+
+    private getTotalUsedStorage = (): number => {
+        let totalUsedStorage: number = 0
+        for (let file of Global.user.fileTracker.files) {
+            totalUsedStorage += file.size
+        }
+        return totalUsedStorage
     }
 
-    private async receiveUpdate() {
-		const settings = ServerConnection.makeSettings();
-        const url = settings.baseUrl + "optumi/list-files";
-		const init: RequestInit = {
-			method: 'GET',
-		};
-		ServerConnection.makeRequest(
-			url,
-			init, 
-			settings
-		).then((response: Response) => {
-			if (this.polling) {
-				// If we are polling, send a new request in 2 seconds
-                if (Global.shouldLogOnPoll) console.log('FunctionPoll (' + new Date().getSeconds() + ')');
-				setTimeout(() => this.receiveUpdate(), 30000);
-			}
-			Global.handleResponse(response);
-			return response.json();
-		}).then((body: any) => {
-			if (body) {
-                this.safeSetState({ files: body.files });
-			}
-        });
+    private async deleteFile(file: FileMetadata) {
+        Global.user.fileTracker.deleteFiles([file]);
     }
 
+    private async downloadFile(file: FileMetadata) {
+		Global.user.fileTracker.downloadFiles(file.path, [file], false);
+    }
+
+    private async deleteDirectory(path: string) {
+        let filesToDelete: FileMetadata[] = []
+        for (let file of Global.user.fileTracker.files) {
+            if (file.path.startsWith(path)) {
+                filesToDelete.push(file)
+            }
+        }
+        Global.user.fileTracker.deleteFiles(filesToDelete);
+    }
+
+    private async downloadDirectory(path: string) {
+        let filesToDownload: FileMetadata[] = []
+        for (let file of Global.user.fileTracker.files) {
+            if (file.path.startsWith(path)) {
+                filesToDownload.push(file)
+            }
+        }
+        Global.user.fileTracker.downloadFiles(path, filesToDownload, false);
+    }
+
+    private handleFilesChange = () => this.forceUpdate();
+
     // Will be called automatically when the component is mounted
 	public componentDidMount = () => {
-        this._isMounted = true;
-        this.polling = true;
-        this.receiveUpdate();
+        Global.user.fileTracker.getFilesChanged().connect(this.handleFilesChange);
 	}
 
 	// Will be called automatically when the component is unmounted
 	public componentWillUnmount = () => {
-        this._isMounted = false;
-        this.polling = false;
+        Global.user.fileTracker.getFilesChanged().disconnect(this.handleFilesChange);
     }
-    
-    private safeSetState = (map: any) => {
-		if (this._isMounted) {
-			let update = false
-			try {
-				for (const key of Object.keys(map)) {
-					if (JSON.stringify(map[key]) !== JSON.stringify((this.state as any)[key])) {
-						update = true
-						break
-					}
-				}
-			} catch (error) {
-				update = true
-			}
-			if (update) {
-				if (Global.shouldLogOnSafeSetState) console.log('SafeSetState (' + new Date().getSeconds() + ')');
-				this.setState(map)
-			} else {
-				if (Global.shouldLogOnSafeSetState) console.log('SuppressedSetState (' + new Date().getSeconds() + ')');
-			}
-		}
-	}
 
     public shouldComponentUpdate = (nextProps: IProps, nextState: IState): boolean => {
         try {
