@@ -75,6 +75,7 @@ class FileProgress {
 export class FileTracker {
 	private polling = false;
     private fileProgress: FileProgress[]
+    private _files: FileMetadata[]
 
 	constructor() {
 		this.polling = true;
@@ -83,6 +84,7 @@ export class FileTracker {
         this.receiveCompressionUpdates(true);
         this.receiveUploadUpdates(true);
         this.receiveDownloadUpdates(true);
+        this.receiveUpdate();
 	}
 
     public get = (name: string): FileProgress[] => {
@@ -114,13 +116,13 @@ export class FileTracker {
         // If we are already uploading this file, ignore the request
         if (this.compressions.filter(x => x.path == metadata.path).length > 0 || this.uploads.filter(x => x.path == metadata.path).length > 0) return;
 
-        const fileNames = [];
+        const paths = [];
         if (metadata.type == 'directory') {
             for (var file of (await FileServerUtils.getRecursiveTree(metadata.path))) {
-                fileNames.push(file);
+                paths.push(file);
             }
         } else {
-            fileNames.push(metadata.path);
+            paths.push(metadata.path);
         }
 
         // Make this unique by adding a timestamp
@@ -132,35 +134,86 @@ export class FileTracker {
 			method: 'POST',
 			body: JSON.stringify({
                 key: metadata.path + key,
-				fileNames: fileNames,
+				paths: paths,
                 compress: Global.user.compressFilesEnabled,
 			}),
 		};
-		ServerConnection.makeRequest(
+        if (Global.user.compressFilesEnabled) this.fileProgress.push(new FileProgress(metadata.path, key, 'compression'));
+        this.fileProgress.push(new FileProgress(metadata.path, key, 'upload'));
+        this._filesChanged.emit(void 0);
+        return ServerConnection.makeRequest(
 			url,
 			init, 
 			settings
 		).then((response: Response) => {
 			Global.handleResponse(response);
 		});
-        if (Global.user.compressFilesEnabled) this.fileProgress.push(new FileProgress(metadata.path, key, 'compression'));
-        this.fileProgress.push(new FileProgress(metadata.path, key, 'upload'));
-        this._filesChanged.emit(void 0);
 	}
 
-    public downloadFiles = async (name: string, files: FileMetadata[], workloadUUID: string, moduleUUID: string, overwrite: boolean) => {
+    public getNotebookOutputFiles = async (name: string, files: FileMetadata[], workloadUUID: string, moduleUUID: string, overwrite: boolean, savePaths: string[] = []) => {
         // If there is an unsigned agreement, do not poll
         if (Global.user != null && Global.user.unsignedAgreement) {
             return;
         }
 
-        const fileNames = [];
+        const addPaths = savePaths.length == 0;
+
+        const paths = [];
         const sizes = [];
         for (var file of files) {
-            fileNames.push(file.path);
+            paths.push(file.path);
+            if (addPaths) savePaths.push(file.path);
             sizes.push(file.size);
         }
         
+        // Make this unique by adding a timestamp
+        const key = new Date().toISOString();
+
+		const settings = ServerConnection.makeSettings();
+		const url = settings.baseUrl + "optumi/get-notebook-output-files";
+		const init: RequestInit = {
+			method: 'POST',
+			body: JSON.stringify({
+				workloadUUID: workloadUUID,
+				moduleUUID: moduleUUID,
+				key: name + key,
+                paths: paths,
+                savePaths: savePaths,
+                sizes: sizes,
+                overwrite: overwrite,
+			}),
+		};
+        this.fileProgress.push(new FileProgress(name, key, 'download'));
+        this._filesChanged.emit(void 0);
+        return ServerConnection.makeRequest(
+			url,
+			init, 
+			settings
+		).then((response: Response) => {
+			Global.handleResponse(response);
+		});
+	}
+
+    public downloadFiles = async (name: string, files: FileMetadata[], overwrite: boolean, savePaths: string[] = []) => {
+        // If there is an unsigned agreement, do not poll
+        if (Global.user != null && Global.user.unsignedAgreement) {
+            return;
+        }
+
+        const addPaths = savePaths.length == 0;
+
+        const paths = [];
+        const sizes = [];
+        const hashes = [];
+        for (var file of files) {
+            if (this.fileExists(file)) {
+                paths.push(file.path);
+                if (addPaths) savePaths.push(file.path);
+                sizes.push(file.size);
+                hashes.push(file.hash || "");
+            }
+        }
+        
         // Make this unique by adding a timestamp
         const key = new Date().toISOString();
 
@@ -169,23 +222,23 @@ export class FileTracker {
 		const init: RequestInit = {
 			method: 'POST',
 			body: JSON.stringify({
-				workloadUUID: workloadUUID,
-				moduleUUID: moduleUUID,
 				key: name + key,
-                files: fileNames,
+                paths: paths,
+                savePaths: savePaths,
                 sizes: sizes,
+                hashes: hashes,
                 overwrite: overwrite,
 			}),
 		};
-		ServerConnection.makeRequest(
+        this.fileProgress.push(new FileProgress(name, key, 'download'));
+        this._filesChanged.emit(void 0);
+        return ServerConnection.makeRequest(
 			url,
 			init, 
 			settings
 		).then((response: Response) => {
 			Global.handleResponse(response);
 		});
-        this.fileProgress.push(new FileProgress(name, key, 'download'));
-        this._filesChanged.emit(void 0);
 	}
 
     private compressionPollDelay = 500;
@@ -215,7 +268,7 @@ export class FileTracker {
                 init, 
                 settings
             ).then((response: Response) => {
-                Global.handleResponse(response);
+                Global.handleResponse(response, true);
                 setTimeout(() => this.receiveCompressionUpdates(), this.compressionPollDelay);
                 if (response.status == 204) {
                     return;
@@ -279,7 +332,7 @@ export class FileTracker {
                 init, 
                 settings
             ).then((response: Response) => {
-                Global.handleResponse(response);
+                Global.handleResponse(response, true);
                 setTimeout(() => this.receiveUploadUpdates(), this.uploadPollDelay);
                 if (response.status == 204) {
                     return;
@@ -341,7 +394,7 @@ export class FileTracker {
                 init, 
                 settings
             ).then((response: Response) => {
-                Global.handleResponse(response);
+                Global.handleResponse(response, true);
                 setTimeout(() => this.receiveDownloadUpdates(), this.downloadPollDelay);
                 if (response.status == 204) {
                     return;
@@ -376,6 +429,145 @@ export class FileTracker {
         }
 	}
 
+    public deleteFiles = async (files: FileMetadata[]) => {
+        // If there is an unsigned agreement, do not poll
+        if (Global.user != null && Global.user.unsignedAgreement) {
+            return;
+        }
+
+        this._files.filter(file => !files.map(x => x.hash + x.path).includes(file.hash + file.path));
+        this._filesChanged.emit();
+
+        const hashes = []
+        const paths = [];
+        for (var metadata of files) {
+            hashes.push(metadata.hash);
+            paths.push(metadata.path);
+        }
+
+		const settings = ServerConnection.makeSettings();
+		const url = settings.baseUrl + "optumi/delete-files";
+		const init: RequestInit = {
+			method: 'POST',
+			body: JSON.stringify({
+				hashes: hashes,
+                paths: paths,
+			}),
+		};
+        return ServerConnection.makeRequest(
+			url,
+			init, 
+			settings
+		).then((response: Response) => {
+			Global.handleResponse(response);
+		});
+	}
+
+    private filesPollDelay = 30000;
+    private receiveUpdate = async () => {
+        if (!this.polling) return;
+        // If there is an unsigned agreement, do not poll
+        if (Global.user != null && Global.user.unsignedAgreement) {
+            setTimeout(() => this.receiveUpdate(), this.filesPollDelay);
+            return;
+        }
+
+		const settings = ServerConnection.makeSettings();
+        const url = settings.baseUrl + "optumi/list-files";
+		const init: RequestInit = {
+			method: 'GET',
+		};
+		ServerConnection.makeRequest(
+			url,
+			init, 
+			settings
+		).then((response: Response) => {
+			if (this.polling) {
+				// If we are polling, send a new request in 30 seconds
+                if (Global.shouldLogOnPoll) console.log('FunctionPoll (' + new Date().getSeconds() + ')');
+				setTimeout(() => this.receiveUpdate(), this.filesPollDelay);
+			}
+			Global.handleResponse(response);
+			return response.json();
+		}).then((body: any) => {
+			if (body) {
+                let newFiles: FileMetadata[] = []
+                if (body.files) {
+                    for (let i = 0; i < body.files.length; i++) {
+                        if (body.files[i] != '') {
+                            newFiles.push({
+                                last_modified: body.filesmod[i],
+                                name: (body.files[i] as string).split('/').pop(),
+                                path: body.files[i],
+                                size: +body.filessize[i],
+                                type: 'file',
+                                hash: body.hashes[i],
+                            } as FileMetadata);
+                        }
+                    }
+                }
+                // const filesToApps = new Map();
+                // const appsToFiles = new Map();
+                // const appTracker = Global.user.appTracker;
+        
+                // // Figure out which apps are associated with which files
+                // for (let metadata of newFiles) {
+                //     const apps = [];
+                //     APP_LOOP:
+                //     for (let app of [...appTracker.activeJobsOrSessions, ...appTracker.finishedJobsOrSessions]) {
+                //         for (let file of app.files) {
+                //             if (file.path === metadata.path) {
+                //                 apps.push(app.uuid);
+                //                 continue APP_LOOP;
+                //             }
+                //         }
+                //         for (let module of app.modules) {
+                //             for (let file of module.files) {
+                //                 if (file.path === metadata.path) {
+                //                     apps.push(app.uuid);
+                //                     continue APP_LOOP;
+                //                 }
+                //             }
+                //         }
+                //     }
+                //     filesToApps.set(metadata, apps);
+                // }
+
+                // // Figure out which files are associated with which apps
+                // for (let app of [...appTracker.activeJobsOrSessions, ...appTracker.finishedJobsOrSessions]) {
+                //     const files = []
+                //     for (let file of app.files) {
+                //         files.push(file)
+                //     }
+                //     for (let module of app.modules) {
+                //         for (let file of module.files) {
+                //             files.push(file);
+                //         }
+                //     }
+                //     appsToFiles.set(app.uuid, files);
+                // }
+
+                this._files = newFiles;
+                this._filesChanged.emit();
+			}
+        }, (error: ServerConnection.ResponseError) => {
+            setTimeout(() => this.receiveUpdate(), this.filesPollDelay);
+        });
+    }
+
+    public get files(): FileMetadata[] {
+        return this._files;
+    }
+
+    public fileExists = (metadata: FileMetadata): boolean => {
+        const match = metadata.hash + metadata.path;
+        for (let file of this._files) {
+            if (file.hash + file.path === match) return true;
+        }
+        return false;
+    }
+
+
     public stopPolling = () => {
         this.polling = false;
     }
