@@ -10,7 +10,7 @@ import { OptionsObject } from "notistack";
 /// NOTE: Look here for possible props: https://iamhosseindhv.com/notistack/api#enqueuesnackbar-options
 
 export const providerOptions = {
-    success: { backgroundColor: '#68da7c!important' },
+    success: { backgroundColor: '#7cdf8e!important' },
     error: { backgroundColor: '#f48f8d!important' },
     warning: { backgroundColor: '#ffba7d!important' },
     info: { backgroundColor: '#10A0F9!important' },
