@@ -64,11 +64,13 @@ export class App {
 	private _preparing: ProgressMessage;
 	private _running: ProgressMessage;
 
+	private _files: FileMetadata[];
+
 	private _timestamp: Date;
 	private _runNum: number;
 
 	constructor(path: string, notebook: any = {}, config: OptumiConfig = new OptumiConfig(), uuid: string = "",
-		initializing: Update[] = [], preparing: Update[] = [], running: Update[] = [], timestamp = new Date(), runNum: number = 0) {		
+		initializing: Update[] = [], preparing: Update[] = [], running: Update[] = [], files: FileMetadata[] = [], timestamp = new Date(), runNum: number = 0) {		
 
 		this._notebook = notebook;
 		this._config = config.copy();
@@ -80,6 +82,8 @@ export class App {
 		this._preparing = new ProgressMessage(Phase.Preparing, preparing);
 		this._running = new ProgressMessage(Phase.Running, running);
 		
+		this._files = files;
+
 		if (this._uuid != "") {
 			this._initializing.appUUID = this._uuid;
 			this._preparing.appUUID = this._uuid;
@@ -122,31 +126,61 @@ export class App {
         for (let i = 0; i < appMap.running.length; i++) {
             running.push(new Update(appMap.running[i], appMap.runningmod[i]));
         }
-		var app: App = new App(appMap.name, JSON.parse(appMap.notebook), new OptumiConfig(JSON.parse(appMap.nbConfig)), appMap.uuid, initializing, preparing, running, new Date(appMap.timestamp), appMap.runNum);
+
+		const inputFiles: FileMetadata[] = [];
+		if (appMap.files) {
+			for (let i = 0; i < appMap.files.length; i++) {
+				if (appMap.files[i] != '') {
+					inputFiles.push({
+						last_modified: appMap.filesmod[i],
+						name: (appMap.files[i] as string).split('/').pop(),
+						path: appMap.files[i],
+						size: +appMap.filessize[i],
+						type: 'file',
+						hash: appMap.hashes[i],
+					} as FileMetadata);
+				}
+			}
+		}
+
+		var app: App = new App(appMap.name, JSON.parse(appMap.notebook), new OptumiConfig(JSON.parse(appMap.nbConfig)), appMap.uuid, initializing, preparing, running, inputFiles, new Date(appMap.timestamp), appMap.runNum);
 		// Add modules
 		for (let module of appMap.modules) {
             const output: Update[] = [];
-            for (let i = 0; i < module.output.length; i++) {
-				output.push(new Update(module.output[i], module.outputmod[i]));
-            }
+			if (module.output) {
+				for (let i = 0; i < module.output.length; i++) {
+					output.push(new Update(module.output[i], module.outputmod[i]));
+				}
+			}
 			const updates: Update[] = [];
-            for (let i = 0; i < module.updates.length; i++) {
-				updates.push(new Update(module.updates[i], module.updatesmod[i]));
-            }
-            const files: FileMetadata[] = [];
-            for (let i = 0; i < module.files.length; i++) {
-                if (module.files[i] != '') {
-                    files.push({
-						last_modified: module.filesmod[i],
-						name: (module.files[i] as string).split('/').pop(),
-						path: module.files[i],
-						size: +module.filessize[i],
-						type: 'file',
-					} as FileMetadata);
-                }
+			if (module.updates) {
+				for (let i = 0; i < module.updates.length; i++) {
+					updates.push(new Update(module.updates[i], module.updatesmod[i]));
+				}
+			}
+            const outputFiles: FileMetadata[] = [];
+			if (module.files) {
+				for (let i = 0; i < module.files.length; i++) {
+					if (module.files[i] != '') {
+						outputFiles.push({
+							last_modified: module.filesmod[i],
+							name: (module.files[i] as string).split('/').pop(),
+							path: module.files[i],
+							size: +module.filessize[i],
+							type: 'file',
+							hash: module.hashes[i],
+						} as FileMetadata);
+					}
+				}
+			}
+			const monitoring: Update[] = [];
+			if (module.monitoring) {
+				for (let i = 0; i < module.monitoring.length; i++) {
+					monitoring.push(new Update(module.monitoring[i], module.monitoringmod[i]));
+				}
 			}
 			var lastPatch: number = 0;
-			if (module.patches != null) {
+			if (module.patches) {
 				for (let i = 0; i < module.patches.length; i++) {
 					const patch = module.patches[i];
 					try {
@@ -158,7 +192,7 @@ export class App {
 					}
 				}
 			}
-			if (module.notebook != null) {
+			if (module.notebook) {
 				const notebook = module.notebook;
 				try {
 					app.updateNotebook(JSON.parse(notebook));
@@ -169,7 +203,7 @@ export class App {
 				}
 			}
 
-			var mod: Module = new Module(module.uuid, module.machine ? Object.setPrototypeOf(module.machine, Machine.prototype) : null, module.token, output, updates, files, lastPatch);
+			var mod: Module = new Module(module.uuid, module.machine ? Object.setPrototypeOf(module.machine, Machine.prototype) : null, module.token, output, updates, outputFiles, monitoring, lastPatch);
 			mod.applyPatch = app.applyPatch;
 			app._modules.push(mod);
 			if (mod.modStatus == Status.RUNNING) {
@@ -310,6 +344,10 @@ export class App {
 		return this._running;
 	}
 
+	get files(): FileMetadata[] {
+		return this._files;
+	}
+
 	get timestamp(): Date {
 		return this._timestamp;
 	}
@@ -527,14 +565,14 @@ export class App {
 			data.requirementsFile =  requirements;
 		}
 
-		data.dataFiles = [];
+		data.paths = [];
 		for (var uploadEntry of uploadFiles) {
 			if (uploadEntry.type == 'directory') {
 				for (var file of (await FileServerUtils.getRecursiveTree(uploadEntry.path))) {
-					data.dataFiles.push(file);
+					data.paths.push(file);
 				}
 			} else {
-				data.dataFiles.push(uploadEntry.path);
+				data.paths.push(uploadEntry.path);
 			}
 		}
 
@@ -559,11 +597,27 @@ export class App {
 			Global.handleResponse(response);
 			// TODO:JJ Take a look at this
 			this._initializing.addUpdate(new Update("stop", ""));
+			return response.json();
 		}, (error: ServerConnection.ResponseError) => {
 			// TODO:JJ Take a look at this
 			this._initializing.addUpdate(new Update('Initialization failed', ""));
 			this._initializing.addUpdate(new Update('error', ""));
 			this._initializing.addUpdate(new Update('stop', ""));
+		}).then((body: any) => {
+			if (body.files) {
+				for (let i = 0; i < body.files.length; i++) {
+					if (body.files[i] != '') {
+						this._files.push({
+							last_modified: body.filesmod[i],
+							name: (body.files[i] as string).split('/').pop(),
+							path: body.files[i],
+							size: +body.filessize[i],
+							type: 'file',
+							hash: body.hashes[i],
+						} as FileMetadata);
+					}
+				}
+			}
 		});
 		this.getLaunchStatus();
 	}
@@ -590,6 +644,8 @@ export class App {
 	adjustMessage(message: string): string {
 		// We will say a session is starting until we can connect to it
 		if (this.interactive && message == 'Running' && !(this.modules.length > 0 && this.modules[0].sessionReady)) return 'Connecting';
+		// We call a running app 'Connected'
+		if (this.interactive && message == 'Running') return 'Connected';
 		// We call a terminated app 'closed'
 		if (this.interactive && message == 'Terminating') return 'Closing';
 		if (this.interactive && message == 'Terminated') return 'Closed';
