@@ -60,6 +60,7 @@ export default class FileServerUtils {
 			const promises: Promise<string[]>[] = [];
 			for (var file of body.content) {
 				if (file.type == 'directory') {
+					// TODO:JJ This is a problem when we have a large number of files
 					promises.push(this.getRecursiveTree(file.path));
 				} else {
 					files.push(file.path);
@@ -74,9 +75,43 @@ export default class FileServerUtils {
 		return files;
 	}
 
-	public static async saveNotebook(path: string, notebook: any): Promise<boolean> {
+	public static async saveNotebook(filePath: string, notebook: any): Promise<boolean> {
+		const splitPath = filePath.split("/");
+		let path = "";
+		for (let i = 0; i < splitPath.length-1; i++) {
+			path += splitPath[i]
+			if (i != splitPath.length-2) {
+				path += "/"
+			}
+		}
+		
+		var builtPath = '';
+		for (let dir of path.split('/')) {
+			builtPath += '/' + dir
+			const settings = ServerConnection.makeSettings();
+			const url = settings.baseUrl + "api/contents" + builtPath;
+			const body = (await ServerConnection.makeRequest(
+				url, 
+				{
+					body: JSON.stringify({
+						'path': builtPath.substring(1),
+						'type': 'directory',
+					}),
+					method: 'PUT'
+				}, 
+				settings
+			).then(response => {
+				if (response.status !== 200 && response.status !== 201) {
+					throw new ServerConnection.ResponseError(response);
+				}
+				return response.json();
+			}));
+
+			if (!body) return false;
+		}
+
 		const settings = ServerConnection.makeSettings();
-		const url = settings.baseUrl + "api/contents/" + path;
+		const url = settings.baseUrl + "api/contents/" + filePath;
 		const response = ServerConnection.makeRequest(
 			url, 
 			{
